# This file is part of Tryton.  The COPYRIGHT file at the top level of
# this repository contains the full copyright notices and license terms.
import datetime as dt
from decimal import Decimal

from trytond.config import config
from trytond.pool import Pool, PoolMeta

from trytond.modules.product import round_price


class Template(metaclass=PoolMeta):
    __name__ = 'product.template'

    @classmethod
    def __setup__(cls):
        super(Template, cls).__setup__()
        new_sel = ('fifo', 'FIFO')
        if new_sel not in cls.cost_price_method.selection:
            cls.cost_price_method._field.selection.append(new_sel)


class Product(metaclass=PoolMeta):
    __name__ = 'product.product'

    def _get_available_fifo_moves(self, date=None, offset=0, limit=None):
        pool = Pool()
        Move = pool.get('stock.move')

        domain = [
            ('product', '=', self.id),
            self._domain_moves_cost(),
            ('from_location.type', '!=', 'storage'),
            ('to_location.type', '=', 'storage'),
            ]
        if not date:
            domain.append(('fifo_quantity_available', '>', 0))
        else:
            domain.append(('effective_date', '<=', date))
        return Move.search(
            domain,
            offset=offset, limit=limit,
            order=[('effective_date', 'DESC'), ('id', 'DESC')])

    def get_fifo_move(self, quantity=0.0, date=None):
        '''
        Return a list of (move, qty) where move is the move to be
        consumed and qty is the quantity (in the product default uom)
        to be consumed on this move. The list contains the "first in"
        moves for the given quantity.
        '''
        pool = Pool()
        Uom = pool.get('product.uom')

        avail_qty = self._get_storage_quantity(date=date)
        if date:
            # On recomputation, we must pretend
            # outgoing moves are not yet done.
            avail_qty += quantity
        fifo_moves = []

        size = config.getint('cache', 'record')

        def moves():
            offset, limit = 0, size
            while True:
                moves = self._get_available_fifo_moves(
                    date=date, offset=offset, limit=limit)
                if not moves:
                    break
                for move in moves:
                    yield move
                offset += size

        for move in moves():
            qty = move.fifo_quantity_available if not date else move.quantity
            qty = Uom.compute_qty(
                move.uom, qty, self.default_uom, round=False)
            avail_qty -= qty

            if avail_qty <= quantity:
                if avail_qty > 0.0:
                    fifo_moves.append(
                        (move, min(qty, quantity - avail_qty)))
                else:
                    fifo_moves.append(
                        (move, min(quantity, qty + avail_qty)))
                    break
        fifo_moves.reverse()
        return fifo_moves

    def recompute_cost_price_fifo(self, start=None):
        pool = Pool()
        Move = pool.get('stock.move')
        Uom = pool.get('product.uom')
        Revision = pool.get('product.cost_price.revision')

        domain = [
            ('product', '=', self.id),
            self._domain_moves_cost(),
            ['OR',
                self._domain_in_moves_cost(),
                self._domain_out_moves_cost(),
                ]
            ]
        if start:
            domain.append(('effective_date', '>=', start))
        moves = Move.search(
                domain, order=[('effective_date', 'ASC'), ('id', 'ASC')])

        _in_moves = Move.search([
                ('product', '=', self.id),
                self._domain_moves_cost(),
                self._domain_in_moves_cost(),
                ], order=[])
        _in_moves = set(m.id for m in _in_moves)

        revisions = Revision.get_for_product(self)

        cost_price = Decimal(0)
        quantity = 0
        if start:
            domain.remove(('effective_date', '>=', start))
            domain.append(('effective_date', '<', start))
            domain.append(self._domain_in_moves_cost())
            prev_moves = Move.search(
                domain,
                order=[('effective_date', 'DESC'), ('id', 'DESC')],
                limit=1)
            if prev_moves:
                move, = prev_moves
                cost_price = move.cost_price
                quantity = self._get_storage_quantity(
                    date=start - dt.timedelta(days=1))
                quantity = Decimal(str(quantity))

        def in_move(move):
            return move.id in _in_moves

        def out_move(move):
            return not in_move(move)

        def compute_fifo_cost_price(quantity, date):
            fifo_moves = self.get_fifo_move(float(quantity), date=date)

            cost_price = Decimal(0)
            consumed_qty = 0
            for move, move_qty in fifo_moves:
                consumed_qty += move_qty
                cost_price += move.get_cost_price() * Decimal(str(move_qty))
            if consumed_qty:
                return round_price(cost_price / Decimal(str(consumed_qty)))

        # For each day, process the incoming moves first
        # in order to keep quantity positive where possible
        # We do not re-browse because we expect only small changes
        moves = sorted(moves, key=lambda m: (
                m.effective_date, out_move(m), m.id))
        current_moves = []
        current_out_qty = 0
        current_cost_price = cost_price
        for move in moves:
            if (current_moves
                    and current_moves[-1].effective_date
                    != move.effective_date):
                Move.write([
                        m for m in filter(in_move, current_moves)
                        if m.cost_price != current_cost_price],
                    dict(cost_price=current_cost_price))

                out_moves = list(filter(out_move, current_moves))
                if out_moves:
                    fifo_cost_price = compute_fifo_cost_price(
                        current_out_qty, current_moves[-1].effective_date)
                    if fifo_cost_price is None:
                        fifo_cost_price = current_cost_price
                    Move.write([
                            m for m in out_moves
                            if m.cost_price != fifo_cost_price],
                        dict(cost_price=fifo_cost_price))
                    if quantity > 0 and quantity + current_out_qty >= 0:
                        cost_price = (
                            ((current_cost_price * (
                                        quantity + current_out_qty))
                                - (fifo_cost_price * current_out_qty))
                            / quantity)
                    else:
                        cost_price = current_cost_price
                    current_cost_price = round_price(cost_price)
                current_moves.clear()
                current_out_qty = 0
            current_moves.append(move)

            cost_price = Revision.apply_up_to(
                revisions, cost_price, move.effective_date)
            qty = Uom.compute_qty(move.uom, move.quantity, self.default_uom)
            qty = Decimal(str(qty))
            if out_move(move):
                qty *= -1
            if in_move(move):
                unit_price = move.get_cost_price(product_cost_price=cost_price)
                if quantity + qty > 0 and quantity >= 0:
                    cost_price = (
                        (cost_price * quantity) + (unit_price * qty)
                        ) / (quantity + qty)
                elif qty > 0:
                    cost_price = unit_price
                current_cost_price = round_price(cost_price)
            elif out_move(move):
                current_out_qty += -qty
            quantity += qty

        Move.write([
                m for m in filter(in_move, current_moves)
                if m.cost_price != current_cost_price],
            dict(cost_price=current_cost_price))

        out_moves = list(filter(out_move, current_moves))
        if out_moves:
            fifo_cost_price = compute_fifo_cost_price(
                current_out_qty, current_moves[-1].effective_date)
            if fifo_cost_price is None:
                fifo_cost_price = current_cost_price
            Move.write([
                    m for m in out_moves
                    if m.cost_price != fifo_cost_price],
                dict(cost_price=fifo_cost_price))
            if quantity:
                cost_price = (
                    ((cost_price * (quantity + current_out_qty))
                        - (fifo_cost_price * current_out_qty))
                    / quantity)
        for revision in revisions:
            cost_price = revision.get_cost_price(cost_price)
        return cost_price



# This file is part of Tryton.  The COPYRIGHT file at the top level of
# this repository contains the full copyright notices and license terms.

try:
    from trytond.modules.product_cost_fifo.tests.test_product_cost_fifo import suite  # noqa: E501
except ImportError:
    from .test_product_cost_fifo import suite

__all__ = ['suite']



# This file is part of Tryton.  The COPYRIGHT file at the top level of
# this repository contains the full copyright notices and license terms.
from decimal import Decimal

from sql import operators, Literal

from trytond.i18n import gettext
from trytond.model import Workflow, ModelView, fields, Check
from trytond.model.exceptions import AccessError
from trytond.pool import Pool, PoolMeta
from trytond.transaction import Transaction

from trytond.modules.product import round_price


class Move(metaclass=PoolMeta):
    __name__ = 'stock.move'
    fifo_quantity = fields.Float(
        'FIFO Quantity',
        help="Quantity used by FIFO.")
    fifo_quantity_available = fields.Function(fields.Float(
            "FIFO Quantity Available",
            help="Quantity available for FIFO"),
        'get_fifo_quantity_available')

    @classmethod
    def __setup__(cls):
        super(Move, cls).__setup__()
        cls._allow_modify_closed_period.add('fifo_quantity')

        t = cls.__table__()
        cls._sql_constraints += [
            ('check_fifo_quantity_out',
                Check(t, t.quantity >= t.fifo_quantity),
                'product_cost_fifo.msg_move_fifo_quantity_greater'),
            ]

    @staticmethod
    def default_fifo_quantity():
        return 0.0

    def get_fifo_quantity_available(self, name):
        return self.quantity - (self.fifo_quantity or 0)

    @classmethod
    def domain_fifo_quantity_available(cls, domain, tables):
        table, _ = tables[None]
        name, operator, value = domain
        field = cls.fifo_quantity_available._field
        Operator = fields.SQL_OPERATORS[operator]
        column = (
            cls.quantity.sql_column(table)
            - cls.fifo_quantity.sql_column(table))
        expression = Operator(column, field._domain_value(operator, value))
        if isinstance(expression, operators.In) and not expression.right:
            expression = Literal(False)
        elif isinstance(expression, operators.NotIn) and not expression.right:
            expression = Literal(True)
        expression = field._domain_add_null(
            column, operator, value, expression)
        return expression

    def _update_fifo_out_product_cost_price(self):
        '''
        Update the product cost price of the given product on the move. Update
        fifo_quantity on the concerned incomming moves. Return the
        cost price for outputing the given product and quantity.
        '''
        pool = Pool()
        Uom = pool.get('product.uom')

        total_qty = Uom.compute_qty(self.uom, self.quantity,
            self.product.default_uom, round=False)

        with Transaction().set_context(company=self.company.id):
            fifo_moves = self.product.get_fifo_move(total_qty)

        cost_price = Decimal("0.0")
        consumed_qty = 0.0
        to_save = []
        for move, move_qty in fifo_moves:
            consumed_qty += move_qty
            cost_price += move.get_cost_price() * Decimal(str(move_qty))

            move_qty = Uom.compute_qty(self.product.default_uom, move_qty,
                    move.uom, round=False)
            move.fifo_quantity = (move.fifo_quantity or 0.0) + move_qty
            # Due to float, the fifo quantity result can exceed the quantity.
            assert move.quantity >= move.fifo_quantity - move.uom.rounding
            move.fifo_quantity = min(move.fifo_quantity, move.quantity)
            to_save.append(move)
        if to_save:
            # TODO save in do method when product change
            self.__class__.save(to_save)

        if consumed_qty:
            cost_price = cost_price / Decimal(str(consumed_qty))
        else:
            cost_price = self.product.get_multivalue(
                'cost_price', **self._cost_price_pattern)

        # Compute average cost price
        average_cost_price = self._compute_product_cost_price(
            'out', product_cost_price=cost_price)

        if cost_price:
            cost_price = round_price(cost_price)
        else:
            cost_price = average_cost_price
        return cost_price, average_cost_price

    def _do(self):
        cost_price = super(Move, self)._do()
        if (self.from_location.type != 'storage'
                and self.to_location.type == 'storage'
                and self.product.cost_price_method == 'fifo'):
            cost_price = self._compute_product_cost_price('in')
        elif (self.to_location.type == 'supplier'
                and self.from_location.type == 'storage'
                and self.product.cost_price_method == 'fifo'):
            cost_price = self._compute_product_cost_price('out')
        elif (self.from_location.type == 'storage'
                and self.to_location.type != 'storage'
                and self.product.cost_price_method == 'fifo'):
            fifo_cost_price, cost_price = (
                self._update_fifo_out_product_cost_price())
            if self.cost_price_required and self.cost_price is None:
                self.cost_price = fifo_cost_price
        return cost_price

    @classmethod
    @ModelView.button
    @Workflow.transition('cancelled')
    def cancel(cls, moves):
        super().cancel(moves)

    @classmethod
    def delete(cls, moves):
        fifo_moves = cls.search([
                ('id', 'in', [m.id for m in moves]),
                ('fifo_quantity', '!=', 0.0),
                ])
        if fifo_moves:
            raise AccessError(
                gettext('product_cost_fifo.msg_move_delete_fifo',
                    move=fifo_moves[0].rec_name))
        super(Move, cls).delete(moves)

    @classmethod
    def copy(cls, moves, default=None):
        if default is None:
            default = {}
        else:
            default = default.copy()
        default.setdefault('fifo_quantity', cls.default_fifo_quantity())
        return super().copy(moves, default=default)
